local _, ns = ...
if not ns.L then ns.L = { } end

local L = setmetatable(ns.L, { __index = function(t, k)
	if not k then return "" end
	local v = tostring(k)
	t[k] = v
	return v
end })

local SharedMedia = LibStub('LibSharedMedia-3.0')

local frame = CreateFrame("Frame", nil, InterfaceOptionsFramePanelContainer)
frame.name = "OhSnap"
frame:Hide()
frame.default = function(self)
	OhSnapDB = OhSnap.Defaults
	print(L["OhSnap: Reset requires relogging or reloading UI. Please do either one of them."])
end

local tempanchor
frame:SetScript("OnShow", function()
	local GAP = 8
	local tekcheck = LibStub("tekKonfig-Checkbox")

	local title, subtitle = LibStub("tekKonfig-Heading").new(frame, "OhSnap", "")

	local showanchor = tekcheck.new(frame, nil, L["Show anchor"], "TOPLEFT", title, "BOTTOMLEFT", -2, -GAP)
	showanchor.tiptext = L["Toggle the text anchor."]
	showanchor:SetChecked(OhSnapDB.showanchor)
	local checksound = showanchor:GetScript("OnClick")
	showanchor:SetScript("OnClick", function(self)
		checksound(self)
		OhSnapDB.showanchor = not OhSnapDB.showanchor
		OhSnap:ToggleAnchor(OhSnapDB.showanchor)
	end)

	local testmode = tekcheck.new(frame, nil, L["Test Mode"], "TOPLEFT", showanchor, "TOPRIGHT", 100, 0)
	testmode.tiptext = L["Toggle the test mode."]
	testmode:SetChecked(OhSnapDB.testmode)
	local checksound = testmode:GetScript("OnClick")
	testmode:SetScript("OnClick", function(self)
		checksound(self)
		OhSnapDB.testmode = not OhSnapDB.testmode
	end)

	for i=1,4 do
		local subtitle = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
		subtitle:SetPoint("TOPLEFT", tempanchor and tempanchor or showanchor, "TOPLEFT", 0, tempanchor and -65 or -35)
		subtitle:SetText(L["Priority "]..i..L[" options:"])

		local enable = tekcheck.new(frame, nil, L["Enable"], "TOPLEFT", subtitle, "BOTTOMLEFT", 0, -GAP)
		enable.tiptext = L["Enable this priority"]
		enable:SetChecked(OhSnapDB[i][3])
		enable:SetHitRectInsets(5, 5, 5, 5)
		local checksound = enable:GetScript("OnClick")
		enable:SetScript("OnClick", function(self)
			checksound(self)
			OhSnapDB[i][3] = not OhSnapDB[i][3]
		end)
		tempanchor = enable

		local dropdown1, dropdown1text, dropdown1container, dropdown1label = LibStub("tekKonfig-Dropdown").new(frame, "", "BOTTOMLEFT", enable, "BOTTOMLEFT", 0, -4)
		dropdown1container:SetHeight(28)
		dropdown1:SetWidth(190)
		dropdown1text:SetText(OhSnapDB[i][1])
		dropdown1.tiptext = L["Select the font style"]

		local function OnClick(self)
			dropdown1text:SetText(self.value)
			OhSnapDB[i][1] = self.value
			OhSnap:Update()
		end

		UIDropDownMenu_Initialize(dropdown1, function()
			local selected, info = OhSnapDB[i][1], UIDropDownMenu_CreateInfo()
			info.func = OnClick

			for k, v in next, SharedMedia:List('font') do
				info.text = v
				info.value = v
				info.checked = v == selected
				UIDropDownMenu_AddButton(info)
			end
		end)

		local dropdown2, dropdown2text, dropdown2container, dropdown2label = LibStub("tekKonfig-Dropdown").new(frame, "", "BOTTOMLEFT", dropdown1, "TOPRIGHT", -5, -4)
		dropdown2container:SetHeight(28)
		dropdown2:SetWidth(120)
		dropdown2text:SetText(L[OhSnapDB[i][2]])
		dropdown2.tiptext = L["Select the font size"]

		local function OnClick(self)
			dropdown2text:SetText(L[self.value])
			OhSnapDB[i][2] = self.value
			OhSnap:Update()
		end

		UIDropDownMenu_Initialize(dropdown2, function()
			local selected, info = OhSnapDB[i][2], UIDropDownMenu_CreateInfo()
			info.func = OnClick

			info.text = L["Huge"]
			info.value = "Huge"
			info.checked = "Huge" == selected
			UIDropDownMenu_AddButton(info)

			info.text = L["Large"]
			info.value = "Large"
			info.checked = "Large" == selected
			UIDropDownMenu_AddButton(info)

			info.text = L["Normal"]
			info.value = "Normal"
			info.checked = "Normal" == selected
			UIDropDownMenu_AddButton(info)

			info.text = L["Small"]
			info.value = "Small"
			info.checked = "Small" == selected
			UIDropDownMenu_AddButton(info)
		end)

		local caster = tekcheck.new(frame, nil, "", "TOPLEFT", dropdown2, "TOPRIGHT", -5, 0)
		caster.tiptext = L["Show caster name"]
		caster:SetChecked(OhSnapDB[i][4])
		caster:SetHitRectInsets(5, 5, 5, 5)
		local checksound = caster:GetScript("OnClick")
		caster:SetScript("OnClick", function(self)
			checksound(self)
			OhSnapDB[i][4] = not OhSnapDB[i][4]
		end)

		local icon = tekcheck.new(frame, nil, "", "TOPLEFT", caster, "TOPRIGHT", 0, 0)
		icon.tiptext = L["Show spell icon"]
		icon:SetChecked(OhSnapDB[i][5])
		icon:SetHitRectInsets(5, 5, 5, 5)
		local checksound = icon:GetScript("OnClick")
		icon:SetScript("OnClick", function(self)
			checksound(self)
			OhSnapDB[i][5] = not OhSnapDB[i][5]
		end)

		local spell = tekcheck.new(frame, nil, "", "TOPLEFT", icon, "TOPRIGHT", 0, 0)
		spell.tiptext = L["Show spell name"]
		spell:SetChecked(OhSnapDB[i][6])
		spell:SetHitRectInsets(5, 5, 5, 5)
		local checksound = spell:GetScript("OnClick")
		spell:SetScript("OnClick", function(self)
			checksound(self)
			OhSnapDB[i][6] = not OhSnapDB[i][6]
		end)

		if i == 1 then
			local target = tekcheck.new(frame, nil, "", "TOPLEFT", spell, "TOPRIGHT", 0, 0)
			target.tiptext = L["Show target name"]
			target:SetChecked(OhSnapDB[i][7])
			target:SetHitRectInsets(5, 5, 5, 5)
			local checksound = target:GetScript("OnClick")
			target:SetScript("OnClick", function(self)
				checksound(self)
				OhSnapDB[i][7] = not OhSnapDB[i][7]
			end)
		else
			local comment = tekcheck.new(frame, nil, "", "TOPLEFT", spell, "TOPRIGHT", 0, 0)
			comment.tiptext = L["Show spell comment"]
			comment:SetChecked(OhSnapDB[i][7])
			comment:SetHitRectInsets(5, 5, 5, 5)
			local checksound = comment:GetScript("OnClick")
			comment:SetScript("OnClick", function(self)
				checksound(self)
				OhSnapDB[i][7] = not OhSnapDB[i][7]
			end)
		end
	end
	frame:SetScript("OnShow", nil)
end)

InterfaceOptions_AddCategory(frame)
LibStub("tekKonfig-AboutPanel").new("OhSnap", "OhSnap")