local OhDB = "1.0"

OhSnap.Defaults = {
	[1] = {"Friz Quadrata TT","Huge",true,true,true,true,true},
	[2] = {"Friz Quadrata TT","Large",true,true,true,false,true},
	[3] = {"Friz Quadrata TT","Normal",true,true,true,false,true},
	[4] = {"Friz Quadrata TT","Small",true,true,true,false,true},
	["dbversion"] = OhDB,
	["showanchor"] = true,
	["testmode"] = false,
	["position"] = {point = "CENTER", x = 0, y = 300},
}
if not OhSnapDB then OhSnapDB = OhSnap.Defaults end

local _, ns = ...
if not ns.L then ns.L = { } end

local L = setmetatable(ns.L, { __index = function(t, k)
	if not k then return "" end
	local v = tostring(k)
	t[k] = v
	return v
end })

local SharedMedia = LibStub('LibSharedMedia-3.0')
local StartTimer = LibStub('AceTimer-3.0')

local messages = {}
local targetMsgs = {}
local rows = {}
local guidmap = {}
local uidcount = 0

-- Automatically create the sub-tables for GUID
local done = setmetatable({}, {__index = function(t,k)
    local new = {}
    rawset(t, k, new)
    return new
end})

-- Create the anchor frame early
local anchor = CreateFrame("Button", nil, UIParent)
anchor:SetScript("OnEvent", function(self, event, ...)
    if self[event] then return self[event](self, event, ...) end
end)

-- Frame we use to SetScript the OnEvent and OnUpdate
local SpellAlertFrame = CreateFrame("Frame")

local function ZoneCheck()
	local instancetype = select(2,GetInstanceInfo())
	if instancetype == "party" or instancetype == "raid" and not OhSnap.Disabled then
		OhSnap:Disable()
	elseif OhSnap.Disabled and not UnitIsDead("player") and not UnitIsGhost("player") then
		OhSnap:Enable()
	end
end

-- We use this frame to enable/disable the addon when player enter instance
local ZoneWatcher = CreateFrame("Frame")
ZoneWatcher:RegisterEvent("ZONE_CHANGED_NEW_AREA")
ZoneWatcher:SetScript("OnEvent", ZoneCheck)

-- Enable events
function OhSnap:Enable()
	anchor:RegisterEvent("PLAYER_TARGET_CHANGED")
	anchor:RegisterEvent("UNIT_AURA")
	anchor:RegisterEvent("PLAYER_ALIVE")
	anchor:RegisterEvent("PLAYER_LOGIN")
	anchor:RegisterEvent("UNIT_SPELLCAST_START")
	anchor:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	SpellAlertFrame:RegisterEvent("UNIT_SPELLCAST_INTERRUPTED")
	SpellAlertFrame:RegisterEvent("UNIT_SPELLCAST_FAILED")
	SpellAlertFrame:RegisterEvent("UNIT_SPELLCAST_FAILED_QUIET")
	SpellAlertFrame:RegisterEvent("UNIT_SPELLCAST_STOP")
	SpellAlertFrame:RegisterEvent("UNIT_SPELLCAST_FAILED_QUIET")
	SpellAlertFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	if OhSnap.Disabled then print(L["OhSnap enabled, you are alive and not in an instance."]) end
	OhSnap.Disabled = nil
end

-- Disable all events
function OhSnap:Disable()
	anchor:UnregisterAllEvents()
	SpellAlertFrame:UnregisterAllEvents()
	if not OhSnap.Disabled then print(L["OhSnap disabled, you have entered an instance."]) end
	OhSnap.Disabled = true
end

local function ShowTooltip(self)
	GameTooltip:SetOwner(self, "ANCHOR_NONE")
	GameTooltip:SetPoint("BOTTOM", self, "TOP")
	GameTooltip:ClearLines()
	GameTooltip:AddLine(L["Doubleclick to open GUI"])
	GameTooltip:Show()
end

local function HideTooltip()
	GameTooltip:Hide()
end

function OhSnap:Initialize()
    anchor:SetFrameStrata("HIGH")
    anchor:SetHeight(30)
    anchor:SetWidth(150)
    anchor:SetBackdrop(GameTooltip:GetBackdrop())
    anchor:SetBackdropColor(0.3, 0.3, 0.3)
    anchor.text = anchor:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
    anchor.text:SetAllPoints(true)
    anchor.text:SetText(L["OhSnap Anchor"])
    anchor:EnableMouse(true)
    anchor:SetMovable(true)
    anchor:Hide()
	anchor:RegisterForDrag("LeftButton")
	anchor:SetScript("OnEnter", ShowTooltip)
	anchor:SetScript("OnLeave", HideTooltip)
	anchor:SetScript("OnDoubleClick", function(self) InterfaceOptionsFrame_OpenToCategory("OhSnap") end)
	anchor:SetScript("OnDragStart", anchor.StartMoving)
	anchor:SetScript("OnDragStop", function(self)
		self:StopMovingOrSizing()
		OhSnapDB.position.point, OhSnapDB.position.x, OhSnapDB.position.y = "BOTTOMLEFT", self:GetLeft(), self:GetBottom()
	end)
end

-- Adds a message to the alert frame, returns a uid
function OhSnap:AddMessage(msg, priority, r, g, b, a, duration, left)
    local entry = {
        msg = msg or L["Empty message"], 
        pri = priority or 1,
        count = uidcount,
        r = r or 1,
        g = g or 1,
        b = b or 1,
        a = a or 1,
        dura = duration or 0,
        left = left or 0,
    }
    uidcount = uidcount + 1
    table.insert(messages, entry)
    self:Update()
    return entry
end

-- Edits a message if it already exists and is being displayed in alert frame
function OhSnap:EditMessage(uid,duration,left)
    for i=1,#messages do
        if messages[i] == uid then
            messages[i].dura = duration
            messages[i].left = left
            self:Update()
            return
        end
    end
end

-- Removes a message from the alert frame
function OhSnap:DelMessage(uid)
    for i=1,#messages do
        if messages[i] == uid then
            table.remove(messages, i)
            self:Update()
            return
        end
    end
end

function OhSnap:Clear()
    table.wipe(messages)
    table.wipe(done)
    table.wipe(targetMsgs)
    SpellAlertFrame:SetScript("OnUpdate",nil)
    self:Update()
end

local function sortFunc(a, b)
    if a.pri == b.pri then
        return a.count < b.count
    else
        return a.pri < b.pri
    end
end

local tempsize = { ["Small"] = {11,"OUTLINE"}, ["Normal"] = {14,"OUTLINE"}, ["Large"] = {18,"OUTLINE"}, ["Huge"] = {24,"THICKOUTLINE"}, }
local function OhSize(size)
	return tempsize[size][1],tempsize[size][2]
end

-- Updates the frame to display
function OhSnap:Update()
	-- Create enough frames, if necessary
	for i=#rows + 1, #messages, 1 do
        local row = CreateFrame("Frame")
        row:SetFrameStrata("HIGH")
        row.text = row:CreateFontString(nil, "OVERLAY")
        row.text:SetPoint("CENTER", 0, 0)
        row.text:SetJustifyH("CENTER")
        rows[i] = row
    end
    -- Sort the frames
    table.sort(messages, sortFunc)
    -- Anchor the frames in order
    for idx,entry in ipairs(messages) do
        local row = rows[idx]
        if idx == 1 then
            row:ClearAllPoints()
            row:SetPoint("TOP", anchor, "BOTTOM", 0, 0)
        else
            row:ClearAllPoints()
            row:SetPoint("TOP", rows[idx-1], "BOTTOM", 0, 0)
        end
        row:SetHeight(20)
        row:SetWidth(250)
		-- local font = OhFont(OhSnapDB[entry.pri][1])
		local font = OhSnapDB[entry.pri][1]
		local size,outline = OhSize(OhSnapDB[entry.pri][2])
		row.text:SetFont(SharedMedia:Fetch('font', font),size,outline)
        local duration = entry.dura
        -- Coloring the time
        local left = floor(entry.left-GetTime())
        local percent
		if left > 0 then
			percent = left / duration
		else
			percent = 0
		end
        local r,g,b = percent > 0.5 and 2 * (1 - percent) or 1, percent > 0.5 and 1 or 2 * percent, 0 
        local color = string.format("|cff%02x%02x%02x", r*255, g*255, b*255)
        local message = (left and left >= 0) and entry.msg.." ("..color..left.."|r"..L["s"]..")" or entry.msg
        row.text:SetText(message)
        row.text:SetTextColor(entry.r, entry.g, entry.b, entry.a)
        row:Show()
    end
    -- Hide any extra frames
    for i=#rows, #messages + 1, -1 do
        rows[i]:Hide()
    end   
end

-- Actually initialize the code
OhSnap:Initialize()
OhSnap:Enable()

local function unitscan(unit)
    -- Debuffs
    local guid = UnitGUID(unit)
    local prio = 4 -- Profitable debuffs
	if OhSnapDB[prio][3] then
		for k,v in pairs(OhSnap.spells[prio]) do
			local spellname = GetSpellInfo(k)
			spellname = tostring(spellname) -- WHAT?
			local targetclass = select(2,UnitClass(unit))
			if v["multi"] then guid = "MULTI" end
			if (v.class and targetclass == v.class) or not v.class then
				-- If the spell is on the given unit, and its not already done
				if ((UnitIsPlayer(unit) and not UnitIsFriend("player", unit)) or OhSnapDB.testmode) and UnitDebuff(unit, spellname) then
					if not done[guid][spellname] then
						local OhName = (v["multi"] and "") or (OhSnapDB[prio][4] and UnitName(unit):match("[^-]+")..": " or "")
						local OhSpellicon = OhSnapDB[prio][5] and " |T"..select(3,UnitDebuff(unit, spellname))..":0|t " or ""
						local OhSpellname = OhSnapDB[prio][6] and "("..spellname..") "  or ""
						local OhDescription = OhSnapDB[prio][7] and v.msg or ""
						local message = OhName..OhSpellicon..OhSpellname..OhDescription
						local uid = OhSnap:AddMessage(message,prio,0,1,0,1,select(6,UnitDebuff(unit,spellname)),select(7,UnitDebuff(unit,spellname)))
						done[guid][spellname] = uid
						if UnitIsUnit(unit, "target") then
							table.insert(targetMsgs, uid)
						end
					end
				elseif done[guid][spellname] then
					local uid = done[guid][spellname]
					OhSnap:DelMessage(uid)
					done[guid][spellname] = nil
				end
			end
		end
	end
    -- Buffs
    for prio=2,3 do -- Noticeable buffs & Annoying buffs
		if OhSnapDB[prio][3] then
			for k,v in pairs(OhSnap.spells[prio]) do
				local spellname = GetSpellInfo(k)
				local targetclass = select(2,UnitClass(unit))
				if v["multi"] then guid = "MULTI" end
				-- Mages have Spellsteal. Let's imagine they can have all the buffs listed :)
				if (v.class and (targetclass == v.class or targetclass == "MAGE")) or not v.class then
					-- If the spell is on the given unit, and its not already done
					if ((UnitIsPlayer(unit) and not UnitIsFriend("player", unit)) or OhSnapDB.testmode) and (spellname ~= nil and UnitAura(unit, spellname)) then
						if not done[guid][spellname] then
							local classcolor = RAID_CLASS_COLORS[select(2,UnitClass(unit))]
							local r,g,b = classcolor.r,classcolor.g,classcolor.b
							local OhName = (v["multi"] and "") or (OhSnapDB[prio][4] and UnitName(unit):match("[^-]+")..": " or "")
							local OhSpellicon = OhSnapDB[prio][5] and "|T"..select(3,UnitAura(unit, spellname))..":0|t " or ""
							local OhSpellname = OhSnapDB[prio][6] and "("..spellname..") " or ""
							local OhDescription = OhSnapDB[prio][7] and v.msg or ""
							local message = OhName..OhSpellicon..OhSpellname..OhDescription
							local uid = OhSnap:AddMessage(message,prio,r,g,b,1,select(6,UnitAura(unit,spellname)),select(7,UnitAura(unit,spellname)))
							done[guid][spellname] = uid
							if UnitIsUnit(unit, "target") then
								 table.insert(targetMsgs, uid)
							end
						else
							-- Message exists but time might have changed
							local uid = done[guid][spellname]
							OhSnap:EditMessage(uid,select(6,UnitAura(unit,spellname)),select(7,UnitAura(unit,spellname)))
						end
					elseif done[guid][spellname] then
						local uid = done[guid][spellname]
						OhSnap:DelMessage(uid)
						done[guid][spellname] = nil
					end
				end
			end
		end
    end
    if done[guid] and not next(done[guid]) then
        done[guid] = nil
    end
    if next(done) then
        if not SpellAlertFrame:GetScript("OnUpdate") then
            SpellAlertFrame:SetScript("OnUpdate", OhSnap.Update)
        end
    else
        if SpellAlertFrame:GetScript("OnUpdate") then
            SpellAlertFrame:SetScript("OnUpdate",nil)
        end
    end
end

function anchor:PLAYER_TARGET_CHANGED(event)
    -- Wipe the table clean when out of arena
    local arena = IsActiveBattlefieldArena()
    if not arena then 
		table.wipe(done)
		for idx,uid in ipairs(targetMsgs) do
			OhSnap:DelMessage(uid)
		end
		table.wipe(targetMsgs)
		if UnitExists("target") and ((UnitIsPlayer("target") and not UnitIsFriend("player", "target")) or OhSnapDB.testmode) then
			unitscan("target")
		end
	end
end

function anchor:UNIT_AURA(event, unit)
    if unit == "target" or unit:match("^arena") then
        unitscan(unit)
    end
end

function anchor:PLAYER_ALIVE()
    if not anchor:IsVisible() then OhSnap:Clear() end
end

function anchor:PLAYER_LOGIN()
	if OhSnapDB.dbversion ~= OhDB then 
		print(L["OhSnap settings cleared."])
		OhSnapDB = OhSnap.Defaults
	end
	anchor:SetPoint(OhSnapDB.position.point, OhSnapDB.position.x, OhSnapDB.position.y)
	ZoneCheck()
    if not anchor:IsVisible() then OhSnap:Clear() end
    OhSnap:ToggleAnchor(OhSnapDB["showanchor"])
end

function anchor:INCOMING_SPELLCAST(event, ...)
    local arena = IsActiveBattlefieldArena()
    -- When we get a unit casting event, store the GUID
    if event == "UNIT_SPELLCAST_START" then
        local unit, spell = ...
        local guid = UnitGUID(unit)
        guidmap[guid] = unit
        return
    end
    -- Use COMBAT_LOG_EVENT_UNFILTERED to actually catch the spellcasts from our non-targets
    if event == "COMBAT_LOG_EVENT_UNFILTERED" then
	local _, cevent, _, sourceGUID, sourceName, _, _, _, destName = ...
        if cevent == "SPELL_CAST_START" then
            local unit = guidmap[sourceGUID]
            if unit and UnitGUID(unit) == sourceGUID and (UnitIsEnemy("player", unit) or OhSnapDB.testmode) then
                -- This is a unit we have an ID for at the moment so grab the target information
                local destName = UnitName(unit .. "target")
                local isUnit = destName and UnitName(destName)
                local spellId, spellName = select(12, ...)
				local spellTexture = select(3, GetSpellInfo(spellId))
                local srcName = sourceName:match("[^-]+")
                local guid = sourceGUID
                if arena and not destName and UnitIsEnemy("player", unit) then
                    -- This is here to ensure we don't skip it
                    destName = L["Unknown"]
                elseif arena and isUnit and (not UnitPlayerOrPetInParty(destName)) and (not UnitIsUnit(destName, "player")) then
                    -- They are casting on someone we don't care about
                    return
                elseif not arena and isUnit and not UnitIsUnit(destName, "player") then
                    -- Ignore anything that isn't targeting us
                    return
                end
                local prio = 1 -- Dangerous spells
				if OhSnapDB[prio][3] then
					for k,v in pairs(OhSnap.spells[prio]) do
						local spellname = GetSpellInfo(k)
						if spellname == spellName then
							if not done[guid][spellname] then
								local class = select(2, UnitClass(unit)) or "PRIEST"
								local classcolor = RAID_CLASS_COLORS[class]
								local r,g,b = classcolor.r, classcolor.g, classcolor.b
								local OhName = OhSnapDB[prio][4] and srcName..": " or ""
								local OhSpellicon = OhSnapDB[prio][5] and "|T"..spellTexture..":0|t " or ""
								local OhSpellname = OhSnapDB[prio][6] and spellName.." " or ""
								-- I hope this will fix the destName = nil -Mind Control error
								local OhDescription = (OhSnapDB[prio][7] and not v.notarget) and (destName and "-> "..destName or "") or ""
								local msg = OhName..OhSpellicon..OhSpellname..OhDescription
								local uid = OhSnap:AddMessage(msg, prio, r, g, b)

								-- Here's the fix for spells that might stick on screen: 
								-- ie. You have no target and someone starts to cast at you -> you target the person and the message sticks on screen.
								local casttime = select(7,GetSpellInfo(spellId))/1000
								StartTimer:ScheduleTimer(function() OhSnap:DelMessage(uid) end, casttime)
								
								done[guid][spellName] = uid
								if targetMsg then 
									table.insert(targetMsgs, uid)
								end
							end
						end
					end
				end
            end
        end
    end
end

anchor.UNIT_SPELLCAST_START = anchor.INCOMING_SPELLCAST
anchor.COMBAT_LOG_EVENT_UNFILTERED = anchor.INCOMING_SPELLCAST

SpellAlertFrame:SetScript("OnEvent", function(self, event, ...)
    local guid, spellname
    if event == "COMBAT_LOG_EVENT_UNFILTERED" then
        local cevent, srcGUID = ...
        if cevent == "SPELL_CAST_SUCCESS" or cevent == "SPELL_CAST_FAILED" then
            spellname = select(13, ...)
            guid = srcGUID
        else
            return
        end
    else
        -- One of the UNIT_ events
        local unit, spell = ...
        guid = UnitGUID(unit)
        spellname = spell
    end

    local uid = done[guid][spellname]
    if uid then 
        OhSnap:DelMessage(uid)
        done[guid][spellname] = nil
    end
end)

function OhSnap:ToggleAnchor(value)
    if value then
        if anchor:IsVisible() then return end
        anchor:Show()
        OhSnapDB["showanchor"] = value
        TestMessage1 = OhSnap:AddMessage("|TInterface\\Icons\\INV_Misc_Bone_HumanSkull_02:0|t " .. L["Dangerous spells"],1,1,0,0)
        TestMessage2 = OhSnap:AddMessage("|TInterface\\Icons\\INV_Misc_Bone_HumanSkull_02:0|t " .. L["Noticeable buffs"],2,1,1,1)
        TestMessage3 = OhSnap:AddMessage("|TInterface\\Icons\\INV_Misc_Bone_HumanSkull_02:0|t " .. L["Annoying buffs"],3,1,1,0)
        TestMessage4 = OhSnap:AddMessage("|TInterface\\Icons\\INV_Misc_Bone_HumanSkull_02:0|t " .. L["Profitable debuffs"],4,0,1,0)
    else
        anchor:Hide()
        OhSnapDB["showanchor"] = value
        OhSnap:DelMessage(TestMessage1)
        OhSnap:DelMessage(TestMessage2)
        OhSnap:DelMessage(TestMessage3)
        OhSnap:DelMessage(TestMessage4)
        TestMessage1 = nil
        TestMessage2 = nil
        TestMessage3 = nil
        TestMessage4 = nil
    end
end

local TestMessage1,TestMessage2,TestMessage3,TestMessage4
SLASH_OhSnap1 = "/ohsnap"
SlashCmdList["OhSnap"] = function(name)
    if name == "hide" then
        OhSnap:ToggleAnchor(false)
    elseif name == "show" then
        OhSnap:ToggleAnchor(true)
    elseif name == "gui" then
        InterfaceOptionsFrame_OpenToCategory("OhSnap")
    elseif name == "clear" then
        OhSnap:Clear()
    elseif not name or name == "" then
        ChatFrame1:AddMessage(L["OhSnap slashcommand:"])
        ChatFrame1:AddMessage(" /ohsnap show - show the anchor")
        ChatFrame1:AddMessage(" /ohsnap hide - hide the anchor")
        ChatFrame1:AddMessage(" /ohsnap gui - open the GUI")
        ChatFrame1:AddMessage(" /ohsnap clear - clears the messages from anchor")
    end
end